/******************************************************************************
 *
 * drvBitBusGpib.c,v (This file is best viewed with a tabwidth of 4)
 *
 ******************************************************************************/

/********************COPYRIGHT NOTIFICATION**********************************
This software is subject to the the conditions described in COPYRIGHT_Combined
and COPYRIGHT_BESSY files included as part of this distribution.
***************************************************************************/

/*
 *        Low level GPIB driver for BitBus controllers
 *
 * Current Author: Marty Kraimer
 * Original Author: John Winans
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <dbDefs.h>
#include <epicsMutex.h>
#include <epicsEvent.h>
#include <drvSup.h>
#include <errlog.h>
#include <drvBitBusInterface.h>

#include <drvHwGpibInterface.h>

#define STATIC static
extern struct drvBitBusEt drvBitBus;

/******************************************************************************
 *
 * This structure is used to hold the hardware-specific information for a
 * single BitBus GPIB link. 
 *
 ******************************************************************************/
struct bbIbLink
{
    struct ibLink head;            /* associated ibLink structure */
    epicsEventId syncSem;
    struct dpvtBitBusHead bbdpvt;
    struct bbIbLink *next;        /* Next BitBus link structure in list */
};

STATIC struct bbIbLink *rootBBLink = 0;    /* Head of bitbus structures */
STATIC epicsMutexId rootBBLock;            /* Mutex for rootBBLink */
int bbIbDebug;

STATIC long bbIbReport(int level)
{
    struct bbIbLink *pbbIbLink = rootBBLink;
    while(pbbIbLink) {
        printf("gpib link %d is bitBus link %d node %d\n",
            (int)pbbIbLink->head.link,
            (int)pbbIbLink->bbdpvt.link,(int)pbbIbLink->bbdpvt.txMsg.node);
        pbbIbLink = pbbIbLink->next;
    }
    return 0;
}

STATIC long bbIbInit(void)
{
    if(rootBBLock) return(0); /* already initialized */
    rootBBLock = epicsMutexMustCreate();
    return(0);
}

STATIC long bbIbGenLink(ibLink *pibLink, int pass)
{
    struct bbIbLink *pbbIbLink = (struct bbIbLink *) pibLink;
    if(pass == 0)/* link is not started yet, do it now */
    {
        return drvGpibLinkStart(&pbbIbLink->head);
    }
    /* nothing to be done when task is started */
    return 0;
}

STATIC long bbIbDestroyLink(ibLink *pibLink)
{
    return 0;
}

STATIC long bbIbReportLink(ibLink *pibLink, int level)
{
    struct bbIbLink *pbbIbLink = (struct bbIbLink *)pibLink;

    printf("  BitBus : link %hd, node %hd",
        pbbIbLink->bbdpvt.link, pbbIbLink->bbdpvt.txMsg.node);
    printf(" ibLink %p dpvtBitBusHead %p\n",
        &pbbIbLink->head,&pbbIbLink->bbdpvt);
    return 0;
}

STATIC long bbIbProbe(ibLink *pibLink)
{
    return 0;
}

STATIC long bbIbRead(    
    ibLink *pibLink,
    int device,
    char *buffer,
    int length,
    int time,
    int eos)
{
    /* The bbIbLink structure starts with the ibLink, so this is 0 */
    struct bbIbLink *pbbIbLink = (struct bbIbLink *) pibLink;
    int bytesRead;

    if(bbIbDebug) errlogPrintf("%s(%p, %d, %p, %d, %d, %d): entered ",
        __FUNCTION__, pibLink, device, buffer, length, time, eos);
    bytesRead = 0;
    pbbIbLink->bbdpvt.rxMsg.data = (unsigned char *) buffer;
    pbbIbLink->bbdpvt.txMsg.cmd = BB_IBCMD_READ_XACT | device;
    pbbIbLink->bbdpvt.txMsg.length = 7;/* send header only */
    pbbIbLink->bbdpvt.rxMsg.cmd = 0;        /* init for the while loop */
    pbbIbLink->bbdpvt.status = BB_OK;
    while (length && (pbbIbLink->bbdpvt.status == BB_OK)
    && (!(pbbIbLink->bbdpvt.rxMsg.cmd & (BB_IBSTAT_EOI | BB_IBSTAT_TMO)))) {
        pbbIbLink->bbdpvt.rxMaxLen =
             (length > BB_MAX_DAT_LEN) ? BB_MAX_DAT_LEN + 7 : length + 7;
        pbbIbLink->bbdpvt.ageLimit = 0;
        if (drvBitBus.qReq(&pbbIbLink->bbdpvt, BB_Q_LOW) != 0)
        {
            pbbIbLink->bbdpvt.status = BB_NONODE;
            return (-1);
        }
        epicsEventMustWait(pbbIbLink->syncSem); 
        if(bbIbDebug) errlogPrintf("%s(): %2.2X >%.13s< driver status 0x%2.2X",
            __FUNCTION__, pbbIbLink->bbdpvt.rxMsg.cmd,
            pbbIbLink->bbdpvt.rxMsg.data, pbbIbLink->bbdpvt.status);
        pbbIbLink->bbdpvt.txMsg.cmd = BB_IBCMD_READ;/* in case have more reading to do */
        pbbIbLink->bbdpvt.rxMsg.data += pbbIbLink->bbdpvt.rxMsg.length - 7;
        length -= pbbIbLink->bbdpvt.rxMsg.length - 7;
        bytesRead += pbbIbLink->bbdpvt.rxMsg.length - 7;
    }
    if ((pbbIbLink->bbdpvt.rxMsg.cmd & BB_IBSTAT_TMO) || (pbbIbLink->bbdpvt.status != BB_OK))
        return (-1);
    else
        return (bytesRead);
}

STATIC long bbIbWrite(pibLink, device, buffer, length, time)
    struct ibLink *pibLink;
    int device;
    char *buffer;
    int length;
    int time;
{
    struct bbIbLink *pbbIbLink = (struct bbIbLink *) pibLink;
    unsigned char dbugBuf[BB_MAX_DAT_LEN + 1];
    unsigned char more2GoCommand;
    unsigned char lastCommand;
    int bytesSent;

    if(bbIbDebug) errlogPrintf("%s(%p, %d, %p, %d, %d): entered ",
        __FUNCTION__, pibLink, device, buffer, length, time);
    bytesSent = length;            /* we either get an error or send them all */
    pbbIbLink->bbdpvt.rxMaxLen = 7;        /* only get the header back */
    pbbIbLink->bbdpvt.txMsg.data = (unsigned char *) buffer;
    pbbIbLink->bbdpvt.rxMsg.cmd = 0;        /* Init for error checking */
    pbbIbLink->bbdpvt.status = BB_OK;

    /* if more than BB_MAX_DAT_LEN bytes */
    more2GoCommand = BB_IBCMD_ADDR_WRITE | device;
    /* if less than BB_MAX_DAT_LEN+1 bytes */
    lastCommand = BB_IBCMD_WRITE_XACT | device;

    while (length && (pbbIbLink->bbdpvt.status == BB_OK)
    && (!(pbbIbLink->bbdpvt.rxMsg.cmd & BB_IBSTAT_TMO))) {
        if (length > BB_MAX_DAT_LEN) {
            pbbIbLink->bbdpvt.txMsg.length = BB_MAX_DAT_LEN + 7;
            pbbIbLink->bbdpvt.txMsg.cmd = more2GoCommand;  /* Write to device */
            length -= BB_MAX_DAT_LEN;    /* Ready for next chunk */
            more2GoCommand = BB_IBCMD_WRITE;
            lastCommand = BB_IBCMD_WRITE_EOI;
        } else {
            pbbIbLink->bbdpvt.txMsg.length = length + 7;
            pbbIbLink->bbdpvt.txMsg.cmd = lastCommand;
            length = 0;  /* This is the last one */
        }
        bcopy((char*)pbbIbLink->bbdpvt.txMsg.data,
            (char*)dbugBuf, pbbIbLink->bbdpvt.txMsg.length - 7);
        dbugBuf[pbbIbLink->bbdpvt.txMsg.length - 7] = '\0';
        if(bbIbDebug) errlogPrintf("%s(): sending %2.2X >%s<",
            __FUNCTION__, pbbIbLink->bbdpvt.txMsg.cmd, dbugBuf);
        pbbIbLink->bbdpvt.ageLimit = 0;
        if (drvBitBus.qReq(&pbbIbLink->bbdpvt, BB_Q_HIGH) != 0)
        {
            pbbIbLink->bbdpvt.status = BB_NONODE;
            return (-1);
        }
        epicsEventMustWait(pbbIbLink->syncSem); 
        if(bbIbDebug) errlogPrintf(" RAC status = 0x%2.2X driver status = 0x%2.2X",
            __FUNCTION__, pbbIbLink->bbdpvt.rxMsg.cmd, pbbIbLink->bbdpvt.status);

        pbbIbLink->bbdpvt.txMsg.data += BB_MAX_DAT_LEN;    /* in case there is more */
    }

    /* All done, check to see if we died due to an error */
    if ((pbbIbLink->bbdpvt.rxMsg.cmd & BB_IBSTAT_TMO) || (pbbIbLink->bbdpvt.status != BB_OK))
        return (-1);
    else
        return (bytesSent);
}

STATIC long bbIbCmd(pibLink, buffer, length)
    struct ibLink *pibLink;
    char *buffer;
    int length;
{
    struct bbIbLink *pbbIbLink = (struct bbIbLink *) pibLink;
    int bytesSent;

    if(bbIbDebug) errlogPrintf("%s(%p, %p, %d): entered ",
        __FUNCTION__, pibLink, buffer, length);
    bytesSent = length;
    pbbIbLink->bbdpvt.rxMaxLen = 7;        /* only get the header back */
    pbbIbLink->bbdpvt.status = BB_OK;    /* prime these for the while loop */
    pbbIbLink->bbdpvt.rxMsg.cmd = 0;
    pbbIbLink->bbdpvt.txMsg.cmd = BB_IBCMD_WRITE_CMD;
    pbbIbLink->bbdpvt.txMsg.data = (unsigned char *) buffer;
    while ((length > BB_MAX_DAT_LEN)
    && (pbbIbLink->bbdpvt.status == BB_OK)
    && (!(pbbIbLink->bbdpvt.rxMsg.cmd & BB_IBSTAT_TMO))) {
        pbbIbLink->bbdpvt.txMsg.length = BB_MAX_DAT_LEN + 7; /* send a chunk */
        pbbIbLink->bbdpvt.ageLimit = 0;
        if (drvBitBus.qReq(&pbbIbLink->bbdpvt, BB_Q_HIGH) != 0)
        {
            pbbIbLink->bbdpvt.status = BB_NONODE;
            return (-1);
        }
        epicsEventMustWait(pbbIbLink->syncSem);  
        length -= BB_MAX_DAT_LEN;    /* ready for next chunk */
        pbbIbLink->bbdpvt.txMsg.data += BB_MAX_DAT_LEN;
    }
    if ((pbbIbLink->bbdpvt.status == BB_OK)
    && (!(pbbIbLink->bbdpvt.rxMsg.cmd & BB_IBSTAT_TMO))) {
        if (epicsEventTryWait(pbbIbLink->syncSem) == epicsEventWaitOK)
        {
            if(bbIbDebug) errlogPrintf("%s(): "
                "able to take the dang sync sem before queueing!", __FUNCTION__);
        }
        pbbIbLink->bbdpvt.txMsg.length = length + 7; /* send the last chunk */
        pbbIbLink->bbdpvt.ageLimit = 0;
        if (drvBitBus.qReq(&pbbIbLink->bbdpvt, BB_Q_HIGH) != 0)
        {
            pbbIbLink->bbdpvt.status = BB_NONODE;
            return (-1);
        }
        epicsEventMustWait(pbbIbLink->syncSem);
/* BUG -- check bitbus response */
    }
    return (bytesSent);
}

STATIC long bbIbIoctl(ibLink *pibLink, int cmd, int val)
{
    unsigned char buf[BB_MAX_DAT_LEN];
    /* The bbIbLink structure starts with the ibLink, so this is 0 */
    struct bbIbLink *pbbIbLink = (struct bbIbLink *) pibLink;

    switch (cmd)
    {
    case IBTMO:                /* set timeout time for next transaction only */
        /* build a TMO message to send to the bug */
        pbbIbLink->bbdpvt.txMsg.length = 7;
        pbbIbLink->bbdpvt.txMsg.cmd = BB_IBCMD_SET_TMO;
        pbbIbLink->bbdpvt.txMsg.data = buf;
        buf[0] = val;
        pbbIbLink->bbdpvt.rxMsg.route = 0;
        pbbIbLink->bbdpvt.rxMaxLen = 7;  /* will only get header back anyway */
        pbbIbLink->bbdpvt.ageLimit = 0;
        /* send it to the bug */
        if (drvBitBus.qReq(&pbbIbLink->bbdpvt, BB_Q_HIGH) != 0)
        {
            pbbIbLink->bbdpvt.status = BB_NONODE;
            return (-1);
        }
        epicsEventMustWait(pbbIbLink->syncSem);
        if ((pbbIbLink->bbdpvt.status == BB_OK)
        && (!(pbbIbLink->bbdpvt.rxMsg.cmd & BB_IBSTAT_TMO)))
            return 0;
        else
            return -1;
        break;
    case IBIFC:                /* send an Interface Clear pulse */
        /* build an IFC message to send to the bug */
        pbbIbLink->bbdpvt.txMsg.length = 7;
        pbbIbLink->bbdpvt.txMsg.cmd = BB_IBCMD_IFC;
        pbbIbLink->bbdpvt.rxMsg.route = 0;
        pbbIbLink->bbdpvt.rxMaxLen = 7;    /* will only get header back */
        pbbIbLink->bbdpvt.ageLimit = 0;
        if (drvBitBus.qReq(&pbbIbLink->bbdpvt, BB_Q_HIGH) != 0)
        {
            pbbIbLink->bbdpvt.status = BB_NONODE;
            return (-1);
        }
        epicsEventMustWait(pbbIbLink->syncSem);
        if ((pbbIbLink->bbdpvt.status == BB_OK)
        && (!(pbbIbLink->bbdpvt.rxMsg.cmd & BB_IBSTAT_TMO)))
            return 0;
        else
            return -1;
        break;
    case IBREN:                /* turn the Remote Enable line on or off */
    case IBGTS:                /* go to standby (ATN off etc... ) */
    case IBGTA:                /* go to active (ATN on etc... ) */
        return 0;
        break;
    default:
        errlogPrintf("%s(%p, %d, %8.8X): invalid command requested\n",
            __FUNCTION__, pibLink, cmd, val);
        return -1;
    }
}

STATIC long bbIbSrqStatus(ibLink *pibLink)
{
    return 0;
}

STATIC long bbIbSrqDisable(ibLink *pibLink)
{
    return 0;
}

STATIC long bbIbSrqEnable(ibLink *pibLink)
{
    return 0;
}

/* The entry table to this driver. */
drvHwGpibEt drvBitBusGpib =
{
    13,
    bbIbReport,
    bbIbInit,
    bbIbGenLink,
    bbIbDestroyLink,
    bbIbReportLink,
    bbIbProbe,
    bbIbRead,
    bbIbWrite,
    bbIbCmd,
    bbIbIoctl,
    bbIbSrqStatus,
    bbIbSrqEnable,
    bbIbSrqDisable
};

/* If records use link type BBGPIB_IO then set
   gpiblink = (bblink+1)*256 + node
*/
int drvBitBusGpibConfig(int gpiblink, int bblink, int node)
{
    struct bbIbLink *pbbIbLink;

    bbIbInit();
    epicsMutexMustLock(rootBBLock);
    /* First check to see this (card, bug) combination already exists */
    for(pbbIbLink=rootBBLink; pbbIbLink; pbbIbLink=pbbIbLink->next)
    {
        if(pbbIbLink->head.link == gpiblink)
        {
            errlogPrintf("drvBbGpibConfig already configured link %d, node %d\n", 
                pbbIbLink->bbdpvt.link, pbbIbLink->bbdpvt.txMsg.node);
            epicsMutexUnlock(rootBBLock);
            return -1;
        }
        if(pbbIbLink->bbdpvt.link==bblink && pbbIbLink->bbdpvt.txMsg.node==node)
        {
            errlogPrintf("link %d node %d already configured as gpib link %d\n",
                pbbIbLink->bbdpvt.link, pbbIbLink->bbdpvt.txMsg.node,
                pbbIbLink->head.link);
            epicsMutexUnlock(rootBBLock);
            return -1;
        }
    }
    epicsMutexUnlock(rootBBLock);

    /* This link is not started yet, initialize all the required stuff */
    pbbIbLink = (struct bbIbLink *) calloc(1,sizeof(struct bbIbLink));
    if(!pbbIbLink) {
        errlogPrintf("Can't allocate pbbIbLink structure\n");
        return (-1);
    }

    pbbIbLink->head.link = gpiblink;
    pbbIbLink->head.drvHwGpib = &drvBitBusGpib;
    pbbIbLink->syncSem = epicsEventMustCreate(epicsEventEmpty);
    callbackSetCallback(0,&pbbIbLink->bbdpvt.callback);
    pbbIbLink->bbdpvt.psyncSem = &pbbIbLink->syncSem;
    pbbIbLink->bbdpvt.link = bblink;
    pbbIbLink->bbdpvt.txMsg.route = BB_STANDARD_TX_ROUTE;
    pbbIbLink->bbdpvt.txMsg.node = node;
    pbbIbLink->bbdpvt.txMsg.tasks = BB_GPIB_TASK;
    if(drvGpibLinkInit(&pbbIbLink->head) != 0) {
        epicsEventDestroy(pbbIbLink->syncSem);
        free(pbbIbLink);
        return -1;
    }
    epicsMutexMustLock(rootBBLock);
    pbbIbLink->next = rootBBLink;
    rootBBLink = pbbIbLink;        /* link the new one into the list */
    epicsMutexUnlock(rootBBLock);
    return 0;
}
